@@ -69,6 +69,7 @@ impl FlatBlockFinder {
     // The thresholds are more lenient to allow for correct grain modeling
     // in extreme cases.
     #[must_use]
+    #[allow(clippy::too_many_lines)]
     pub fn run(&self, plane: &Plane<u8>) -> (Vec<u8>, usize) {
         const TRACE_THRESHOLD: f64 = 0.15f64 / BLOCK_SIZE_SQUARED as f64;
         const RATIO_THRESHOLD: f64 = 1.25f64;
@@ -112,19 +113,21 @@ impl FlatBlockFinder {
                 );
                 for yi in 1..(BLOCK_SIZE - 1) {
                     for xi in 1..(BLOCK_SIZE - 1) {
-                        let gx = (block_result[yi * BLOCK_SIZE + xi + 1]
-                            - block_result[yi * BLOCK_SIZE + xi - 1])
-                            / 2f64;
-                        let gy = (block_result[yi * BLOCK_SIZE + xi + BLOCK_SIZE]
-                            - block_result[yi * BLOCK_SIZE + xi - BLOCK_SIZE])
-                            / 2f64;
-                        gxx += gx * gx;
-                        gxy += gx * gy;
-                        gyy += gy * gy;
+                        // SAFETY: We know the size of `block_result` and that we cannot exceed the bounds of it
+                        unsafe {
+                            let result_ptr = block_result.as_ptr().add(yi * BLOCK_SIZE + xi);
 
-                        let block_val = block_result[yi * BLOCK_SIZE + xi];
-                        mean += block_val;
-                        var += block_val * block_val;
+                            let gx = (*result_ptr.add(1) - *result_ptr.sub(1)) / 2f64;
+                            let gy =
+                                (*result_ptr.add(BLOCK_SIZE) - *result_ptr.sub(BLOCK_SIZE)) / 2f64;
+                            gxx += gx * gx;
+                            gxy += gx * gy;
+                            gyy += gy * gy;
+
+                            let block_val = *result_ptr;
+                            mean += block_val;
+                            var += block_val * block_val;
+                        }
                     }
                 }
                 let block_size_norm_factor = (BLOCK_SIZE - 2).pow(2) as f64;
@@ -134,11 +137,11 @@ impl FlatBlockFinder {
                 gxx /= block_size_norm_factor;
                 gxy /= block_size_norm_factor;
                 gyy /= block_size_norm_factor;
-                var = var / block_size_norm_factor - mean.powi(2);
+                var = mean.mul_add(-mean, var / block_size_norm_factor);
 
                 let trace = gxx + gyy;
-                let det = gxx * gyy - gxy.powi(2);
-                let e_sub = (trace.powi(2) - 4f64 * det).max(0.).sqrt();
+                let det = gxx.mul_add(gyy, -gxy.powi(2));
+                let e_sub = (trace.mul_add(trace, -4f64 * det)).max(0.).sqrt();
                 let e1 = (trace + e_sub) / 2.0f64;
                 let e2 = (trace - e_sub) / 2.0f64;
                 // Spectral norm
@@ -159,11 +162,15 @@ impl FlatBlockFinder {
                 // clamp the value to [-25.0, 100.0] to prevent overflow
                 let sum_weights = clamp(sum_weights, -25.0f64, 100.0f64);
                 let score = (1.0f64 / (1.0f64 + (-sum_weights).exp())) as f32;
-                flat_blocks[by * num_blocks_w + bx] = if is_flat { 255 } else { 0 };
-                scores[by * num_blocks_w + bx] = IndexAndScore {
-                    score: if var > VAR_THRESHOLD { score } else { 0f32 },
-                    index: by * num_blocks_w + bx,
-                };
+                // SAFETY: We know the size of `flat_blocks` and `scores` and that we cannot exceed the bounds of it
+                unsafe {
+                    let index = by * num_blocks_w + bx;
+                    *flat_blocks.get_unchecked_mut(index) = if is_flat { 255 } else { 0 };
+                    *scores.get_unchecked_mut(index) = IndexAndScore {
+                        score: if var > VAR_THRESHOLD { score } else { 0f32 },
+                        index,
+                    };
+                }
                 if is_flat {
                     num_flat += 1;
                 }
@@ -171,14 +178,18 @@ impl FlatBlockFinder {
         }
 
         scores.sort_unstable_by(|a, b| a.score.partial_cmp(&b.score).expect("Shouldn't be NaN"));
-        let top_nth_percentile = num_blocks * 90 / 100;
-        let score_threshold = scores[top_nth_percentile].score;
-        for score in &scores {
-            if score.score >= score_threshold {
-                if flat_blocks[score.index] == 0 {
-                    num_flat += 1;
+        // SAFETY: We know the size of `flat_blocks` and `scores` and that we cannot exceed the bounds of it
+        unsafe {
+            let top_nth_percentile = num_blocks * 90 / 100;
+            let score_threshold = scores.get_unchecked(top_nth_percentile).score;
+            for score in &scores {
+                if score.score >= score_threshold {
+                    let block_ref = flat_blocks.get_unchecked_mut(score.index);
+                    if *block_ref == 0 {
+                        num_flat += 1;
+                    }
+                    *block_ref |= 1;
                 }
-                flat_blocks[score.index] |= 1;
             }
         }
 
@@ -195,13 +206,19 @@ impl FlatBlockFinder {
     ) {
         let mut plane_coords = [0f64; LOW_POLY_NUM_PARAMS];
         let mut a_t_a_inv_b = [0f64; LOW_POLY_NUM_PARAMS];
+        let plane_origin = plane.data_origin();
 
         for yi in 0..BLOCK_SIZE {
             let y = clamp(offset_y + yi, 0, plane.cfg.height - 1);
             for xi in 0..BLOCK_SIZE {
                 let x = clamp(offset_x + xi, 0, plane.cfg.width - 1);
-                block_result[yi * BLOCK_SIZE + xi] =
-                    f64::from(plane.data_origin()[y * plane.cfg.stride + x]) / BLOCK_NORMALIZATION;
+                // SAFETY: We know the bounds of the plane data and `block_result`
+                // and do not exceed them.
+                unsafe {
+                    *block_result.get_unchecked_mut(yi * BLOCK_SIZE + xi) =
+                        f64::from(*plane_origin.get_unchecked(y * plane.cfg.stride + x))
+                            / BLOCK_NORMALIZATION;
+                }
             }
         }
 
@@ -230,8 +247,8 @@ impl FlatBlockFinder {
             1,
         );
 
-        for i in 0..BLOCK_SIZE_SQUARED {
-            block_result[i] -= plane_result[i];
+        for (block_res, plane_res) in block_result.iter_mut().zip(plane_result.iter()) {
+            *block_res -= *plane_res;
         }
     }
 }
@@ -760,33 +777,29 @@ impl NoiseModel {
             for bx in 0..num_blocks_w {
                 // SAFETY: We know the indexes we provide do not overflow the data bounds
                 unsafe {
+                    let flat_block_ptr = flat_blocks.as_ptr().add(by * num_blocks_w + bx);
                     let x_o = bx * block_w;
-                    if *flat_blocks.get_unchecked(by * num_blocks_w + bx) == 0 {
+                    if *flat_block_ptr == 0 {
                         continue;
                     }
-                    let y_start =
-                        if by > 0 && *flat_blocks.get_unchecked((by - 1) * num_blocks_w + bx) > 0 {
-                            0
+                    let y_start = if by > 0 && *flat_block_ptr.sub(num_blocks_w) > 0 {
+                        0
+                    } else {
+                        NOISE_MODEL_LAG
+                    };
+                    let x_start = if bx > 0 && *flat_block_ptr.sub(1) > 0 {
+                        0
+                    } else {
+                        NOISE_MODEL_LAG
+                    };
+                    let y_end = ((frame_dims.1 >> dec.1) - by * block_h).min(block_h);
+                    let x_end = ((frame_dims.0 >> dec.0) - bx * block_w - NOISE_MODEL_LAG).min(
+                        if bx + 1 < num_blocks_w && *flat_block_ptr.add(1) > 0 {
+                            block_w
                         } else {
-                            NOISE_MODEL_LAG
-                        };
-                    let x_start =
-                        if bx > 0 && *flat_blocks.get_unchecked(by * num_blocks_w + bx - 1) > 0 {
-                            0
-                        } else {
-                            NOISE_MODEL_LAG
-                        };
-                    let y_end = ((frame_dims.1 >> source.cfg.ydec) - by * block_h).min(block_h);
-                    let x_end =
-                        ((frame_dims.0 >> source.cfg.xdec) - bx * block_w - NOISE_MODEL_LAG).min(
-                            if bx + 1 < num_blocks_w
-                                && *flat_blocks.get_unchecked(by * num_blocks_w + bx + 1) > 0
-                            {
-                                block_w
-                            } else {
-                                block_w - NOISE_MODEL_LAG
-                            },
-                        );
+                            block_w - NOISE_MODEL_LAG
+                        },
+                    );
                     for y in y_start..y_end {
                         for x in x_start..x_end {
                             let val = extract_ar_row(
@@ -889,7 +902,7 @@ impl NoiseModel {
                     // But don't allow fully correlated noise (hence the max), since the
                     // synthesis cannot model it.
                     let uncorr_std = (noise_var / 16f64)
-                        .max(noise_var - (corr * luma_strength).powi(2))
+                        .max((corr * luma_strength).mul_add(-(corr * luma_strength), noise_var))
                         .sqrt();
                     let adjusted_strength = uncorr_std / noise_gain;
                     self.latest_state[channel]
@@ -987,7 +1000,7 @@ impl StrengthSolver {
         let bin = self.get_bin_index(block_mean);
         let bin_i0 = bin.floor() as usize;
         let bin_i1 = (self.num_bins - 1).min(bin_i0 + 1);
-        let a = bin as f64 - bin_i0 as f64;
+        let a = bin - bin_i0 as f64;
         let n = self.num_bins;
         let eqns = &mut self.eqns;
         eqns.a[bin_i0 * n + bin_i0] += (1f64 - a).powi(2);
